/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "a_star.h"

#include <algorithm>
#include <cstring>
#include <cassert>
#include <climits>
#include <unordered_map>
#include <unordered_set>
#include <vector>
#include <tuple>
#include <set>


/*
 * returns the cost of traversing from the start of one route to the end of
 * another. This functions uses memoization
 */
int AStarSearcher::cost(const Route & src, const Route & tgt)
{
	if (src.r_num_ == tgt.r_num_) {
		return src.path.size();
	}
	if (cost_mem_[src.r_num_].count(tgt.r_num_) == 0) {
		int p_cost = cost(src.path.back(), tgt.path.front(), &Mesh::edge_cost);
		cost_mem_[src.r_num_][tgt.r_num_] = p_cost;
		return p_cost;
	} else {
		return cost_mem_[src.r_num_][tgt.r_num_];
	}
}


/*
 * Returns the cost of traveling from one Node to another
 */
int AStarSearcher::cost(const Node & src, const Node & tgt, int (Mesh::*cost_func)(Edge) const)
{	
	const int bl = mesh_.border_length;
	std::unordered_map<int, int> gScore = std::unordered_map<int, int>();
	std::unordered_map<int, int> fScore = std::unordered_map<int, int>();
	std::unordered_set<int> closedSet = std::unordered_set<int>();
	// For efficiency the first value is the fScore
	std::set<int> openSet = std::set<int>();
	int s = cotos(src.x, src.y, bl);
	int t = cotos(tgt.x, tgt.y, bl);
	int c = s;
	openSet.insert(c);
	fScore[c] = heuristic(src,tgt);
	gScore[c] = 0;
	while (not openSet.empty()) {
		c = *openSet.begin();
		for (auto n: openSet) {
			if (fScore[c] > fScore[n]) {
				c = n;
			}
		}
		if (c == t) {
			return gScore[c];
		}
		openSet.erase(c);
		closedSet.insert(c);
		std::vector<Node> neighbors;
		Node current = stoco(c,bl);
		mesh_.get_neighbors(neighbors, current);
		std::random_shuffle(neighbors.begin(), neighbors.end());
		for (auto neighbor: neighbors) {
			int n = cotos(neighbor.x, neighbor.y, bl);
			if (closedSet.find(n) != closedSet.end()) continue;
			int t_gscore = gScore[c] + (mesh_.*cost_func)(Edge(current, neighbor));
			auto r = openSet.insert(n);
			if (!r.second) {
				if (t_gscore > gScore[n]) {
					continue;
				}
			}
			gScore[n] = t_gscore;
			fScore[n] = t_gscore + heuristic(stoco(n, bl), tgt);

		}

	}	
	return -1;
/*	const int bl = mesh_.border_length;
	int gScore[bl][bl];
	int fScore[bl][bl];
	bool closedSet[bl][bl];
	std::set<Node>  openSet;
	openSet.insert(src);
	memset(closedSet, false, closedSet[0][0] * bl * bl);
	memset(gScore, INT_MAX, sizeof(gScore[0][0]) * bl * bl );
	memset(fScore, INT_MAX, sizeof(fScore[0][0]) * bl * bl );
	gScore[src.x][src.y] = 0;
	fScore[src.x][src.y] = heuristic(src, tgt);
	while (not openSet.empty() ) {
		auto c = openSet.begin();
		for (auto n = ++openSet.begin(); n != openSet.end(); ++n) {
			if (fScore[(*c).x][(*c).y] > fScore[(*n).x][(*n).y]) {
				c = n;
			}
		}
		Node current = *c;
		if (current == tgt) return gScore[current.x][current.y];
		openSet.erase(c);
		closedSet[current.x][current.y] = true;
		std::vector<Node> neighbors;
		mesh_.get_neighbors(neighbors, current);
		for (auto neighbor: neighbors) {
			if (closedSet[neighbor.x][neighbor.y]) continue;
			int t_gscore = gScore[current.x][current.y] + mesh_.edge_cost(Edge(current, neighbor));
			if (openSet.find(neighbor) == openSet.end()) {
				openSet.insert(neighbor);
			} else if (t_gscore >= gScore[neighbor.x][neighbor.y]) {
				continue;
			}
			gScore[neighbor.x][neighbor.y] = t_gscore;
			fScore[neighbor.x][neighbor.y] = t_gscore + heuristic(neighbor, tgt);
			
		}
	}
	return -1;
	*/
}


/*
 * Returns the cost of traveling from on node to another and stores the shortest
 * path in dest
 */
int AStarSearcher::path(std::vector<Node> & dest, const Node & src, const Node & tgt)
{ 
	return path(dest, src, tgt, &Mesh::edge_cost);
}


/*
 * Returns the cost of traveling from on node to another and stores the shortest
 * path in dest allows for a cost funtion to be provided
 */
int AStarSearcher::path(std::vector<Node> & dest, const Node & src, const Node & tgt, int (Mesh::*cost_func)(Edge) const)
{
	const int bl = mesh_.border_length;
	std::unordered_map<int, int> gScore = std::unordered_map<int, int>();
	std::unordered_map<int, int> fScore = std::unordered_map<int, int>();
	std::unordered_set<int> closedSet = std::unordered_set<int>();
	// For efficiency the first value is the fScore
	std::set<int> openSet = std::set<int>();
	std::unordered_map<int, int> cameFrom = std::unordered_map<int, int>();
	int s = cotos(src.x, src.y, bl);
	int t = cotos(tgt.x, tgt.y, bl);
	int c = s;
	openSet.insert(c);
	fScore[c] = heuristic(src,tgt);
	gScore[c] = 0;
	while (not openSet.empty()) {
		c = *openSet.begin();
		for (auto n: openSet) {
			if (fScore[c] > fScore[n]) {
				c = n;
			}
		}
		if (c == t) {
			int cost = 0;
			Node p = stoco(t, bl);
			while (c != s) {
				dest.push_back(stoco(c, bl));
				c = cameFrom[c];
				Node cur = stoco(c,bl);
				cost += mesh_.edge_cost(Edge(p,cur));
				p = cur;
			}
			dest.push_back(stoco(c, bl));
			std::reverse(dest.begin(), dest.end());
			return cost;
		}
		openSet.erase(c);
		closedSet.insert(c);
		std::vector<Node> neighbors;
		Node current = stoco(c,bl);
		mesh_.get_neighbors(neighbors, current);
		std::random_shuffle(neighbors.begin(), neighbors.end());
		for (auto neighbor: neighbors) {
			int n = cotos(neighbor.x, neighbor.y, bl);
			if (closedSet.find(n) != closedSet.end()) continue;
			int t_gscore = gScore[c] + (mesh_.*cost_func)(Edge(current, neighbor));
			auto r = openSet.insert(n);
			if (!r.second) {
				if (t_gscore > gScore[n]) {
					continue;
				}
			}
			cameFrom[n] = c;
			gScore[n] = t_gscore;
			fScore[n] = t_gscore + heuristic(stoco(n, bl), tgt);

		}

	}	
	return -1;
}


/*
 * Used to maintain internal references
 */
int AStarSearcher::cotos(int x, int y, int bl)
{
	return x*bl + y;
}


/*
 * Used to maintain internal references
 */
Node AStarSearcher::stoco(int c, int bl)
{
	return Node(c/bl, c%bl);
}

/*
 * A* Huristic. Using max of the hirzontal or vertical distance because diagonal
 * edges are still only a cost of BASE_EDGE_COST TODO CHANGE THIS
 */
int AStarSearcher::heuristic(const Node & src, const Node & tgt)
{
	return std::max(abs(src.x - tgt.x), abs(src.y - tgt.y)) * BASE_EDGE_COST;
}
