/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "city.h"
#include "a_star.h"

#include <cmath>
#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <ostream>
#include <ctime>
#include <utility>
#include <iterator>
#include <vector>
#include <string>
#include <iomanip>
#include <cstring>

#ifdef NUM_THREADS
#include <thread>
#include <mutex>
#endif

int MAX_START_TIME = 2400;
int MAX_DEAD_HEADS =  5;
float BASE_BLOOM_DISPERSAL_RATE = 15;
float WAIT_COST = 1;
float TRAVEL_COST = 2;
std::string ROUTE_METHOD = "RADIAL";
std::string TIME_METHOD = "CIRCADIAN";

City::City(int edge_size):	mesh_(edge_size), searcher(mesh_){}
City::~City(){}


/* Adds a increase in cost in nodes at a random point */
void City::add_random_cost_bloom(int num_blooms)
{
	for (int i = 0; i < num_blooms; ++i) {
		Node centre = Node();
		centre.x = rand() % mesh_.border_length;
		centre.y = rand() % mesh_.border_length;
		int edge_cost = rand()%(MAX_EDGE_COST-BASE_EDGE_COST)+BASE_EDGE_COST;
		int dispersal_rate = (MAX_EDGE_COST-BASE_EDGE_COST)*
				float((rand()%5)+ BASE_BLOOM_DISPERSAL_RATE)/100;
		if (dispersal_rate < 1) dispersal_rate = 1;
		mesh_.add_cost_bloom(centre, edge_cost, dispersal_rate);
	}
}


/* Adds diagonal paths across the */
void City::add_random_diagonal_path(int num_diagonals)
{
	for (int i=0;i<num_diagonals; ++i) {
		Node n = Node(rand() % mesh_.border_length, rand() % mesh_.border_length);
		add_diagonal_path(n);
	}
}



/* Removes nodes */
void City::remove_extra_nodes(int desired_node_count)
{
	assert(desired_node_count < mesh_.num_nodes());
	std::set<Node> remove_nodes = std::set<Node>();
	select_nodes_alg(remove_nodes, mesh_.num_nodes() - desired_node_count);
	for(auto node: remove_nodes) {
		mesh_.remove_node(node);
	}
}


/* Removes edges */
void City::remove_extra_edges(int desired_edge_count)
{
	std::set<Edge> remove_edges = std::set<Edge>();
	select_edges_alg(remove_edges, mesh_.num_edges() - desired_edge_count);
	for(auto edge: remove_edges) {
		mesh_.remove_edge(edge);
	}
}


/* Creates bus routes */
void City::create_routes( int desired_route_count, int desired_runs_per_route)
{
	if ((unsigned int) desired_route_count <= routes_.size()) {
	std::cerr << "There are already " << routes_.size() << " routes" << std::endl;
		exit(0);
	}
	RouteTable new_routes = std::vector<Route>();
	create_routes_alg( new_routes, desired_route_count - routes_.size());
	for( auto & r: new_routes) {
		create_start_time_alg(r.start_time, desired_runs_per_route);
	}
	auto m_iter_b = std::make_move_iterator(new_routes.begin());
	auto m_iter_e = std::make_move_iterator(new_routes.end());
	routes_.insert( routes_.end(), m_iter_b, m_iter_e);
}


/* 
 * Writes the data to the stream 
 * 
 * This is the most inefficient part of the software but there isn't an easy way
 * around it 
 * */
void City::write_test(std::ostream& out)
{
	int indx = 0;
	int costs = 0;
	int cost;
	int rn = 0;
	int rm = routes_.size();
	for(auto & r: routes_) {
		std::cout << '\r' << ++rn << '/' << rm << " pass 1/3          ";
		std::cout.flush();
		int num_indxs = r.start_time.size();
		r.indx.reserve(num_indxs);
		std::generate_n(std::back_inserter(r.indx), num_indxs, 
			[&indx]() { return ++indx; });
	}
	rn = 0;
	int rn2 = 0;
#ifndef NUM_THREADS
	for (Route & r: routes_) {
		++rn;
		rm = 0;
		for (Route & o: routes_) {
			std::cout << '\r' << rn << '/' << rm << ' ' << ++rn2 << '/' << rm << " pass 2/3          ";
			std::cout.flush();
			if ((cost = searcher.cost(r,o)) == -1) {	continue;	}
			cost += r.run_time;
			int found = MAX_DEAD_HEADS;
			for (int st: r.start_time) {
				for (int ot: o.start_time) {
					if (st+cost <= ot) {
						if (found) {
							++costs;
							--found;
						} else { break;	}
					}
				}
				if (not found) { break; }
			}
		}
	}
#else
	int retvals[NUM_THREADS];
	int stats[NUM_THREADS];
	std::memset(retvals, 0, sizeof(int) * NUM_THREADS);
	std::memset(stats, 0, sizeof(int) * NUM_THREADS);
	std::thread thread_list[NUM_THREADS];
	for (int i = 0; i < NUM_THREADS; ++i) {
		thread_list[i] = std::thread(&City::count_costs, this, retvals, i, stats);
	}
	for (int i = 0; i < NUM_THREADS; ++i) {
		thread_list[i].join();
	}
	costs = 0;
	for (int i = 0; i < NUM_THREADS; ++i) {
		costs += retvals[i];
	}
	std::cout << '\r' << std::string(NUM_THREADS * 10+20, ' ') << '\r';
	std::cout.flush();
#endif
	out << indx << ' ' << costs << std::endl;
	for(Route & r: routes_) {
		for(int t: r.start_time) {
			out << t << ' ' << t+r.run_time << std::endl;
			indx--;
		}
	}
	rn = 0;
	rn2 = 0;
	for (Route & r: routes_) {
		++rn;
		for (Route & o: routes_) {
			std::cout << '\r' << rn << '/' << rm << ' ' << ++rn2 << '/' << rm << " pass 3/3          ";
			std::cout.flush();
			if ((cost = searcher.cost(r,o)) == -1) { continue; }
			int found = MAX_DEAD_HEADS;
			int r_cost = cost + r.run_time;
			for (unsigned int i = 0; i < r.start_time.size(); ++i ) {
				int st =  r.start_time[i];
				for (unsigned int j = 0; j < o.start_time.size(); ++j ) {
					int ot = o.start_time[j];
					if (st + r_cost <= ot) {
							if (found) {
							--costs;
							--found;
							int wait_time = ot - (st +r_cost);
							int dead_head_cost = TRAVEL_COST * cost +  WAIT_COST * wait_time;
							out << r.indx[i] << ' ' << o.indx[j] << ' ' << dead_head_cost << std::endl;
						} else { break; }
					}
				}
				if (not found) { break; }
			}
		}
	}
	std::cout << '\r' << std::string(20, ' ') << '\r';
	std::cout.flush();
}

#ifdef NUM_THREADS
std::mutex print_mutex;
void print_status(int * r, int rm)
{
	rm = rm/NUM_THREADS;
	if( print_mutex.try_lock()) {
		std::cout <<'\r';
		for (int i = 0; i < NUM_THREADS; ++i) {
			std::cout << "T:" << i << ' ' ;
			if (r[i] < rm) {
				std::cout << r[i] << '/' << rm << ' ';
			} else {
				std::cout << "done ";
			}
		}
		std::cout << " pass 2/3                  ";
		print_mutex.unlock();
	}
}

void City::count_costs( int * c, int thread_id, int * stats)
{
	for (int i = thread_id; i < routes_.size(); i += NUM_THREADS) {
		Route r = routes_[i];
		print_status(stats, routes_.size());
		++stats[thread_id];
		for (Route & o: routes_) {
			int cost;
			if ((cost = searcher.cost(r,o)) == -1) {	continue;	}
			cost += r.run_time;
			int found = MAX_DEAD_HEADS;
			for (int st: r.start_time) {
				for (int ot: o.start_time) {
					if (st+cost <= ot) {
						if (found) {
							++(c[thread_id]);
							--found;
						} else { break;	}
					}
				}
				if (not found) { break; }
			}
		}
	}
	print_status(stats, routes_.size());
}
#endif

/* Generates a map of all of the bus routes */
void City::print_bus_map(std::ostream& out) const
{
	Mesh m = Mesh(mesh_.border_length, 0);
	for (auto r: routes_) {
		for (auto n: r.path) {
			m.add_node(n);
		}
		Node p = r.path.front();
		for (auto c: r.path) {
			if (p == c) continue;
			m.add_edge(Edge(p, c));
			p = c;
		}
	}
	out << m;
}

/*** 
 *
 * Private 
 *
 ***/

/* Adds a diagonal path */
void City::add_diagonal_path(Node center)
{
	mesh_.remove_node(center);
	Node n;
	int xc [4] = {1,1,-1,-1};
	int yc [4] = {-1,1,1,-1};
	for (int i = 0; i < 4; ++i) {
		int x_mod = xc[i];
		int y_mod = yc[i];
		n = center;
		n.x += x_mod;
		n.y += y_mod;
		add_directional_path(n, x_mod, y_mod);
	}
}


/* Adds a directional path */
void City::add_directional_path( Node start, int x_change, int y_change)
{
	Node n1, n2;
	n1 = start;
	n2 = n1;
	n2.x += x_change;
	n2.y += y_change;
	while ( mesh_.is_valid_node(n2)) {
		if ( mesh_.is_node(n1) and mesh_.is_node(n2) ) {
			mesh_.add_edge(Edge(n1,n2));
		}
		n1.x += x_change;
		n1.y += y_change;
		n2.x += x_change;
		n2.y += y_change;
	}
}


/* Moves a node in a direction */
void City::move_node(Node& n, int dir)
{
	switch(dir) {
	case 0:
		n.y -= 1;
		return;
	case 1:
		n.x +=1;
		n.y -=1;
		return;
	case 2:
		n.x +=1;
		return;
	case 3:
		n.x +=1;
		n.y +=1;
		return;
	case 4:
		n.y +=1;
		return;
	case 5:
		n.x -=1;
		n.y +=1;
		return;
	case 6:
		n.x -=1;
		return;
	case 7:
		n.x -=1;
		n.y -=1;
	}
}


/*** 
 *
 * Algorithm selectors 
 *
 ***/
void City::select_nodes_alg( std::set<Node>& dest, int num_nodes )
{	select_nodes_random( dest, num_nodes ); }

void City::select_edges_alg( std::set<Edge>& dest, int num_edges )
{ select_edges_random( dest, num_edges ); }

void City::create_routes_alg( std::vector<Route>& dest, int num_routes)
{ 
	if (ROUTE_METHOD == "RADIAL") {
		create_routes_radial( dest, num_routes);
	} else if (ROUTE_METHOD == "RANDOM") {
		create_routes_random( dest, num_routes);
	} else {
		std::cout << "Invalid route generation method using default";
		std::cout << "(RADIAL)" << std::endl;
		create_routes_radial( dest, num_routes);
	}
}

void City::create_start_time_alg( std::vector<int>& dest, int num_routes)
{ 
	if (TIME_METHOD == "CIRCADIAN") {
		create_start_time_circadian( dest, num_routes); 
	} else if (TIME_METHOD == "RANDOM") {
		create_start_time_random( dest, num_routes);
	} else {
		std::cout << "Invalid start time generatation methout using default";
		std::cout << "(CIRCADIAN)" << std::endl;
		create_start_time_circadian( dest, num_routes); 
	}
}


/*** 
 *
 * Algorithms
 *
 ***/
void City::select_nodes_random(std::set<Node>& dest, int num_nodes)
{
	assert(num_nodes < mesh_.num_nodes());
	while(num_nodes > 0) {
		Node n = Node(rand() % mesh_.border_length, rand() % mesh_.border_length);
		if(mesh_.is_node(n) and dest.find(n) == dest.end() ) {
			dest.insert(n);
			num_nodes--;
		}
	}
}


void City::select_edges_random(std::set<Edge>& dest, int num_edges)
{
	assert(num_edges < mesh_.num_edges());
	while(num_edges > 0) {
		Node n = Node(rand() % mesh_.border_length, rand() % mesh_.border_length);
		if (mesh_.is_node(n)) {
			int dir = rand() % 8;
			Node n2 = n;
			move_node(n2, dir);
			if (mesh_.is_valid_node(n2)) {
				Edge e = Edge((Node) n, (Node) n2);
				if (mesh_.is_edge(e) and dest.find(e) == dest.end()) {
					dest.insert(e);
					num_edges--;
					n = n2;
				}
			}
		}
	}
}


void City::create_routes_random(std::vector<Route>& dest, int num_routes)
{
	for(; num_routes > 0; --num_routes) {
		int length = (mesh_.border_length/4 + rand()%(mesh_.border_length/2));
		Node nl = Node(rand() % mesh_.border_length, rand() % mesh_.border_length);
		while(!mesh_.is_node(nl)) {
			nl.x = rand() % mesh_.border_length;
			nl.y = rand() % mesh_.border_length;
		}
		Node n = nl; 
		dest.emplace_back();
		dest.back().path.push_back(nl);
		length--;
		while (length > 0 and mesh_.num_neighbors(nl) > 1) {
			int dir = rand() % 8;
			n = nl;
			move_node(n, dir);
			if (mesh_.is_valid_node(n)) {
			auto p = ++dest.back().path.rbegin();
			if (mesh_.is_node(n) and *p != n){
					dest.back().path.push_back(n);
					move_node(nl, dir);
					length--;
				}
			}
		}
		dest.back().run_time = dest.back().path.size();
	}
}

/* Addes a cost to reduce bus route overlap */
void City::add_route_cost(Route & route)
{
	Node & p = route.path.front();
	for (auto n: route.path) {
		if (p != n) {
			mesh_.add_overlap_cost(Edge(p,n));
			p = n;
		}
	}
}


/*
 * Generates bus routes that radiate outwards from the centre Node
 */
void City::create_routes_radial(std::vector<Route>& dest, int num_routes)
{

	int bl = mesh_.border_length;
	std::vector<Node> n;  
	Node centre;
	do {
			double m = bl;
			int x = rand() % int(m*0.25) + int(m*0.375);
			int y = rand() % int(m*0.25) + int(m*0.375);
			centre.x = x;
			centre.y = y;
		
		mesh_.get_neighbors(n, centre);
	} while (n.size() == 0); 
	int spaceing = bl * 4 / num_routes;
	if (not spaceing) {
		spaceing = 1;
	}

		int n_d = 0;
		std::cout << ++n_d << '/' << num_routes;
		std::cout.flush();

	int i;
	for( i = 0; i < bl; i += spaceing) {
		int xmod = rand() % 3 -1;
		int ymod = rand() % 3 -1;
		Node start(centre.x + xmod, centre.y + ymod);
		Node end(i, 0);
		Route n_route = Route();
		n_route.run_time = searcher.path(n_route.path, start, end, &Mesh::edge_and_overlap_cost);
		if (n_route.run_time > 0) {
			if (rand()%2) std::reverse(n_route.path.begin(), n_route.path.end());
			dest.push_back(n_route);
			add_route_cost(n_route);
		}
		std::cout << '\r' << ++n_d << '/' << num_routes;
		std::cout.flush();
	}
	for( i = i % bl; i < bl; i += spaceing) {
		int xmod = rand() % 3 -1;
		int ymod = rand() % 3 -1;
		Node start(centre.x + xmod, centre.y + ymod);
		Node end(0, i);
		Route n_route = Route();
		n_route.run_time = searcher.path(n_route.path, start, end, &Mesh::edge_and_overlap_cost);
		if (n_route.run_time > 0 ) {
			if (rand()%2) std::reverse(n_route.path.begin(), n_route.path.end());
			dest.push_back(n_route);
			add_route_cost(n_route);
		}
		std::cout << '\r' << ++n_d << '/' << num_routes;
		std::cout.flush();
	}
	for( i = bl - i % bl; i >= 0; i -= spaceing) {
		int xmod = rand() % 3 -1;
		int ymod = rand() % 3 -1;
		Node start(centre.x + xmod, centre.y + ymod);
		Node end(i, bl-1);
		Route n_route = Route();
		n_route.run_time = searcher.path(n_route.path, start, end, &Mesh::edge_and_overlap_cost);
		if (n_route.run_time > 0) {
			if (rand()%2) std::reverse(n_route.path.begin(), n_route.path.end());
			dest.push_back(n_route);
			add_route_cost(n_route);
		}
	}
	for( i = bl - i % bl; i >= 0; i -= spaceing) {
		int xmod = rand() % 3 -1;
		int ymod = rand() % 3 -1;
		Node start(centre.x + xmod, centre.y + ymod);
		Node end(bl-1, i);
		Route n_route = Route();
		n_route.run_time = searcher.path(n_route.path, start, end, &Mesh::edge_and_overlap_cost);
		if (n_route.run_time > 0) {
			if (rand()%2) std::reverse(n_route.path.begin(), n_route.path.end());
			dest.push_back(n_route);
			add_route_cost(n_route);
		}
		std::cout << '\r' << ++n_d << '/' << num_routes;
		std::cout.flush();
	}
	std::cout << '\r';
}


void City::create_start_time_random( std::vector<int>& dest, int num_times)
{
	for(;num_times > 0; --num_times) {
		dest.push_back(rand()%MAX_START_TIME);
	}
	std::sort(dest.begin(), dest.end());
}


/*
 * Generates a list of start time to match a circadian cycle
 */
void City::create_start_time_circadian( std::vector<int>& dest, int num_times)
{
	const int PEAK_MAX = num_times/8 + 1;
	std::default_random_engine gen;
	const float block [12][3] = { 
				{ 4.0/16.0,      0.5/24.0,        6.0/24.0}, //0600-0630
        {16.0/16.0,      1.0/24.0,        6.5/24.0}, //0630-0730
        {14.0/16.0,      0.5/24.0,        7.5/24.0}, //0730-0800
        {11.0/16.0,      0.5/24.0,        8.0/24.0}, //0800-0830
        {13.0/16.0,      1.0/24.0,        8.5/24.0}, //0830-0930
        { 7.0/16.0,      3.5/24.0,        9.5/24.0}, //0930-1300
        { 5.0/16.0,      2.0/24.0,       13.0/24.0}, //1300-1500
        {10.0/16.0,      1.5/24.0,       15.0/24.0}, //1500-1630
        { 8.0/16.0,      1.5/24.0,       16.5/24.0}, //1630-1800
        { 5.0/16.0,      1.5/24.0,       18.0/24.0}, //1800-1930
        { 3.0/16.0,      0.5/24.0,       19.5/24.0}, //1930-2000
        { 1.0/16.0,      2.0/24.0,       20.0/24.0}};//2000-2200
	for(int i = 0; i < 12; ++i) {
		std::normal_distribution<float> distr(block[i][0] * PEAK_MAX * block[i][1]*24, 1.0);
		int t_in_block = distr(gen);
		int start = block[i][2]*MAX_START_TIME;
		int inc = block[i][1] / t_in_block*MAX_START_TIME;
		for (int j = 0; j < t_in_block; ++j) {
			int n_start = start + inc*j + ((double)rand()/(RAND_MAX)-0.5)*inc;
			dest.push_back(n_start);
		}
	}
	std::sort(dest.begin(), dest.end());
}

/***
 * Print Function
 ***/
std::ostream& operator<<(std::ostream& out, City const& c)
{
	out << c.mesh_ << std::endl;
	out << "Routes:" << std::endl;
	for (auto r : c.routes_) {
		out << r << std::endl << std::endl;
	}
	c.print_bus_map(out);
	return out;
}

