/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "config_parser.h"
#include "mesh.h"
#include "city.h"
#include "gen_input.h"

#include <algorithm>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <cstring>

using namespace std;
/*
 * Generates a default config file with comments on how to fill it out
 */
void ConfigParser::generate_template(std::ostream& out) const
{

	out << "# Comments are placed in " << endl;

	out << "# Save file name for the generated test case" << endl;
	out << "TEST_FILE: test.in" << endl << endl;

	out << "# Save file for a rempresentation of the city" << endl;
	out << "# Does not print city if left blank." << endl;
	out << "# Prints to stdout if stdout is entered" << endl;
	out << "CITY_FILE: test.city" << endl << endl;
	
	out << "# Config save file. Saves a copy of the config to a file" << endl;
	out << "# including the seed used. This allows for the city to" << endl;
	out << "# be reproducable. Does not save if left blank" << endl;
	out << "CITY_CONFIG_FILE: test.cfg" << endl << endl;

	out << "# Sets the seed for the random number generator." << endl;
	out << "# If set to -1 then the current time is used as a seed" << endl;
	out << "SEED: -1" << endl << endl;

	out << "# Number of nodes on an edge" << endl;
	out << "EDGE_SIZE: 30" << endl << endl;

	out << "# Percentage of node to keep. Setting this number too low" << endl;
	out << "# can cause large number of disconections." << endl;
	out << "MESH_DENSITY: 0.80" << endl << endl;

	out << "# Percentage of edges to keep. Setting this number too low" << endl;
	out << "# can cause large number of disconnections." << endl;
	out << "MESH_CONNECTIVITY: 0.80" << endl << endl;

	out << "# Number of diagonals to create." << endl;
	out << "# Diagonals are paths that run diagonally accros the city" << endl;
	out << "NUM_DIAGONALS: 10" << endl << endl;

	out << "# Number of bus routes to create." << endl;
	out << "# This is the apporoximate number of bus routes to create" << endl;
	out << "# due to randomness functionality there is often a few less" << endl;
	out << "NUM_ROUTES: 30" << endl << endl;

	out << "# Number of times to run a bus on any route" << endl;
	out << "# This is an approximate number of times to run any bus route" << endl;
	out << "# over the course of the timeframe" << endl;
	out << "NUM_RUNS: 52" << endl << endl;
	
	out << "# The number of max number of dead head routes from one start" << endl;
	out << "# time to another start time of the same route." << endl;
	out << "MAX_DEAD_HEADS: 5" << endl << endl;

	out << "# Cost blooms are areas that have an increased cost to simulate" << endl;
	out << "# random increases in trafic in areas" << endl << endl;
	out << "# Controls number of cost blooms generated" << endl;
	out << "NUM_COST_BLOOMS: 100" << endl << endl;

	out << "# Controls the range of costs for any edge" << endl;
	out << "BASE_EDGE_COST: 5" << endl;
	out << "MAX_EDGE_COST: 10" << endl << endl;

	out << "# A cost to overlaping but routes to promote unique paths." << endl;
	out << "OVERLAP_COST: 1" << endl << endl;

	out << "# Used to calculate the dead head route cost."               << endl;
	out << "# Equation:"                                                 << endl;
	out << "#		cost = wait_time * WAIT_COST + tavel_distance * TRAVEL_COST" << endl;
	out << "# wait time is time spent idle waiting for the start time of"<< endl;
	out << "# the new route to start. Travel distance is the cost of" << endl;
	out << "# traveling along the edges." << endl;
	out << "WAIT_COST = 1.0" << endl;
	out << "TRAVEL_COST = 2.0" << endl << endl;

	out << "# Used to control the range of time for bus routes to start." << endl;
	out << "# Generated bus routes will have a start time " << endl;
	out << "# from 0 - MAX_START_TIME" << endl;
	out << "MAX_START_TIME: 2400" << endl << endl;

	out << "# Controls the method for generating routes patterns." << endl;
	out << "# Options:" << endl;
	out << "#     RADIAL: Picks a central node and routes spread out" << endl;
	out << "#             from this node to the edges" << endl;
	out << "#			RANDOM: Randomly selects start points and then selects" << endl;
	out << "#             random edges" << endl;
	out << "ROUTE_METHOD: RADIAL" << endl << endl;
	
	out << "# Controls the method for generating start times for bus" << endl;
	out << "# routes." << endl;
	out << "# Options:" << endl;
	out << "#     CIRCADIAN: The start times are generated to follow a" << endl;
	out << "#                circadian cycle. This will generate more" << endl;
	out << "#                routes at peak times" << endl;
	out << "#     RANDOM: This will generate start times at random times" << endl;
	out << "TIME_METHOD: CIRCADIAN" << endl << endl;
}


/* Writes out the config settings to the stream provided */
void ConfigParser::write_config(std::ostream& out) const
{
	auto t = time(nullptr);
	auto tm = localtime(&t);

	out ;
	out << "# Generated on " << asctime(tm) << endl;

	out << "TEST_FILE: " << TEST_FILE << endl;

	out << "CITY_FILE: "<< CITY_FILE << endl;
	
	out << "CITY_CONFIG_FILE: " << CITY_CONFIG_FILE << endl;

	out << "SEED: " << SEED << endl;

	out << "EDGE_SIZE: " << EDGE_SIZE << endl;

	out << setiosflags(ios::showpoint) << "MESH_DENSITY: " << MESH_DENSITY << endl;

	out << setiosflags(ios::showpoint) << "MESH_CONNECTIVITY: " << MESH_CONNECTIVITY << endl;

	out << "NUM_DIAGONALS: " << NUM_DIAGONALS << endl;

	out << "NUM_ROUTES: " << NUM_ROUTES << endl;

	out << "NUM_RUNS: " << NUM_RUNS << endl;

	out << "MAX_DEAD_HEADS: " << MAX_DEAD_HEADS << endl;

	out << "NUM_COST_BLOOMS: " << NUM_COST_BLOOMS << endl;

	out << "BASE_EDGE_COST: " << BASE_EDGE_COST << endl;

	out << "MAX_EDGE_COST: " << MAX_EDGE_COST << endl;

	out << "OVERLAP_COST: " << OVERLAP_COST << endl;

	out << setiosflags(ios::showpoint) << "WAIT_COST: " << WAIT_COST << endl;

	out << setiosflags(ios::showpoint) << "TRAVEL_COST: " << TRAVEL_COST << endl;

	out << "MAX_START_TIME: " << MAX_START_TIME << endl;

	out << "ROUTE_METHOD: " << ROUTE_METHOD << endl;
	
	out << "TIME_METHOD: " << TIME_METHOD << endl;

}

/* Reads in a config file and parses the values */
void ConfigParser::read_config(istream& is)
{
	string line;
	while(getline(is, line)) {
		istringstream is_line(line);
		string key;
		is_line >> ws;
		if (is_line.peek() == '#') {
			continue;
		} else if (getline(is_line, key, ':')) {
			string value;
			is_line >> ws;
			if (getline(is_line, value)) {
				set_option(key, value);
				cout << "Found option [" << key << "] set to [" << value << "]" << endl;
			}
		}
	}
}

string string_keys[] = {"TEST_FILE", "CITY_FILE", "CITY_CONFIG_FILE",
	"ROUTE_METHOD", "TIME_METHOD"};
string int_keys[] = {"SEED", "EDGE_SIZE", "NUM_DIAGONALS", "NUM_ROUTES",
	"NUM_RUNS", "MAX_DEAD_HEADS", "NUM_COST_BLOOMS", "BASE_EDGE_COST",
	"MAX_EDGE_COST" ,"OVERLAP_COST", "MAX_START_TIME"};
string float_keys[] = {"MESH_DENSITY", "MESH_CONNECTIVITY", "WAIT_COST",
	"TRAVEL_COST"};
/* Parses an option and sets the global value to the value parsed */
void ConfigParser::set_option(string key, string value)
{
	if (find(begin(string_keys), end(string_keys), key) != end(string_keys)) {
		set_option_string(key, value);
	} else if (find(begin(int_keys), end(int_keys), key) != end(int_keys)) {
		set_option_int(key, value);
	} else if (find(begin(float_keys), end(float_keys), key) != end(float_keys)) {
		set_option_float(key, value);
	} else {
		cout << "Key [" << key << "] with value [" << value << "] not valid" << endl;
	}
}

void ConfigParser::set_option_string(string key, string value)
{
	int p = distance(begin(string_keys), find(begin(string_keys), end(string_keys), key));
	switch (p) {
		case (0): TEST_FILE = value; return;
		case (1): CITY_FILE = value; return;
		case (2): CITY_CONFIG_FILE = value; return;
		case (3): ROUTE_METHOD = value; return;
		case (4): TIME_METHOD = value; return;
		default: exit(1);
	}
}
void ConfigParser::set_option_int(string key, string value)
{
	int p = distance(begin(int_keys), find(begin(int_keys), end(int_keys), key));
	int v = stoi(value);
	switch (p) {
		case (0):		 SEED = v; return;
		case (1):		 EDGE_SIZE = v; return;
		case (2):		 NUM_DIAGONALS = v; return;
		case (3):		 NUM_ROUTES = v; return;
		case (4):		 NUM_RUNS = v; return;
		case (5):		 MAX_DEAD_HEADS = v; return;
		case (6):		 NUM_COST_BLOOMS = v; return;
		case (7):		 BASE_EDGE_COST = v; return;
		case (8):		 MAX_EDGE_COST = v; return;
		case (9):		 OVERLAP_COST = v; return;
		case (10):   MAX_START_TIME = v; return;
		default: exit(1);
	}
}
void ConfigParser::set_option_float(string key, string value)
{
	int p = distance(begin(float_keys), find(begin(float_keys), end(float_keys), key));
	float v = stof(value);
	switch (p) {
		case (0): MESH_DENSITY = v; return;
		case (1): MESH_CONNECTIVITY = v; return;
		case (2): WAIT_COST = v; return;
		case (3): TRAVEL_COST = v; return;
		default: exit(1);
	}
}
