/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "node.h"
#include "mesh.h"
#include "city.h"
#include "config_parser.h"

#include <iostream>
#include <fstream>
#include <string>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <unistd.h>
#include <getopt.h>

using namespace std;

string TEST_FILE        = "test.in";
string CITY_FILE        = "stdout";
string CITY_CONFIG_FILE = "test.cfg";
string CITY_CONFIG_FILE_IN = "";
int SEED                = -1;
int EDGE_SIZE           = 30;
float MESH_DENSITY      = 0.8;
float MESH_CONNECTIVITY = 0.8;
int NUM_DIAGONALS       = 10;
int NUM_ROUTES          = 30;
int NUM_RUNS            = 52;
int NUM_COST_BLOOMS         = 100;


int main(int argc, char* argv[]) {
	int a;
	/* Parsing command line arguments */
	int option_index = 0;
	static struct option long_options[] = {
		{"test_file", required_argument, 0, 't' },
		{"city_file", required_argument, 0, 'c' },
		{"icfg_file", required_argument, 0, 'i' },
		{"ocfg_file", required_argument, 0, 'o' },
		{"gen_template", optional_argument,    0, 'g' },
		{"help",      no_argument,       0, 'h' },
		{0,           0,                 0, 0   }};
		const char* ostring = "t:c:i:o:g::h";
	while ((a = getopt_long(argc, argv, ostring, long_options, &option_index)) != -1) {
		switch(a) {
		case 't':
			TEST_FILE = optarg;
			break;
		case 'c':
			CITY_FILE = optarg;
			break;
		case 'i':
			CITY_CONFIG_FILE_IN = optarg;
			break;
		case 'o':
			CITY_CONFIG_FILE = optarg;
			break;
		case 'g':
			{
				ofstream file;
				if (optarg != NULL) {
					file.open(optarg);
				} else {
					file.open("default.cfg");
				}
				ConfigParser cfg;
				cfg.generate_template(file);
				file.close();
				return 0;
			}
		case 'h':
			cout << "Usage: " << argv[0] << " [options]" << endl;
			cout << "Options:" << endl;
			cout << "  -t, --test_file=FILE\t\tStore testcase in FILE" << endl;
			cout << "  -c, --city_file=FILE\t\tStore a visual representation of the city in FILE" << endl;
			cout << "  -i, --icfg_file=FILE\t\tLoad the configuration of the city from FILE" << endl;
			cout << "  -o, --ocfg_file=FILE\t\tStore the configuration of the city in FILE" << endl;
			cout << "  -g, --gen_template[=FILE]\tGenerates a default config file" << endl;

			cout << "  -h \t\t\t\tPrint Help information" << endl;
			return 0;

		case '?':
			cout << "Invalid argument" << endl;
			cout << "See --help for options" << endl;
			return 1;
		case ':':
			cout << "Missing required argument" << endl;
			cout << "See --help for options" << endl;
			return 1;
		}	
	}

	/* Load the config if a file is supplied */
	if (CITY_CONFIG_FILE_IN.length()) {
		ConfigParser cfg;
		ifstream file;
		file.open(CITY_CONFIG_FILE_IN);
		cfg.read_config(file);
		file.close();
	}

	if (SEED == -1) {
		SEED = time(0);
		cout << "A Seed of -1 was provided the seed used is [" << SEED << ']' << endl;
	}
	srand(time(0));

	/* Save the config if a config save is supplied */
	if (CITY_CONFIG_FILE.length()) {
		cout << "Writing configuration to " << CITY_CONFIG_FILE << endl;
		ofstream file;
		file.open(CITY_CONFIG_FILE);
		ConfigParser cfg;
		cfg.write_config(file);
		file.close();
		cout << "Closing file " << CITY_CONFIG_FILE << endl;
	}

	cout << "Generating base city" << endl;
	City c = City(EDGE_SIZE);
	int	node_max = EDGE_SIZE*EDGE_SIZE;
	cout << "Adding cost blooms" << endl;
	c.add_random_cost_bloom(NUM_COST_BLOOMS);

	cout << "Adding diagonals" << endl;
	c.add_random_diagonal_path(NUM_DIAGONALS);

	cout << "Removing extra nodes" << endl;
	c.remove_extra_nodes(MESH_DENSITY * node_max);

	cout << "Removing extra edges" << endl;
	c.remove_extra_edges(node_max * 4 * MESH_CONNECTIVITY);

	cout << "Creating routes" << endl;
	c.create_routes(NUM_ROUTES, NUM_RUNS);

	cout << "Writing tests to " << TEST_FILE << endl;
	if (TEST_FILE.compare("stdout") == 0) {
		c.write_test(cout);
	} else {
		ofstream file;
		file.open(TEST_FILE);
		c.write_test(file); 
		file.close(); 
		cout << "Closing file" << TEST_FILE << endl; 
	} 
	if (CITY_FILE.length()) { 
		cout << "Writing city to " << CITY_FILE << endl;
		if (CITY_FILE.compare("stdout") == 0) { 
			cout << c; 
		} else { 
			ofstream file; 
			file.open(CITY_FILE); 
			file << c; file.close(); 
			cout << "Closing file " << CITY_FILE << endl; 
		} 
	} 
	return 0;
}
