/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "mesh.h"

#include <iostream>
#include <cassert>
#include <cstdlib>

/* This is here just because I'm lazy */
enum dir { N, NE, E, SE, S, SW, W, NW };


int BASE_EDGE_COST = 5;
int MAX_EDGE_COST =  10;
int OVERLAP_COST = 1;

/* Constructor */
Mesh::Mesh(int border_length, int base):
	border_length(border_length),
	num_nodes_(border_length*border_length),
	num_edges_( (border_length-1) * border_length * 2)
{
	nodes = new bool *[border_length];
	edges = new std::bitset<8> *[border_length];
	edge_costs = new int ** [border_length];
	overlap_costs = new int ** [border_length];

	for (int i=0; i < border_length; ++i) {
		nodes[i] = new bool[border_length];
		std::fill_n(nodes[i], border_length, base!=0);
		/* init all edges to be available */
		edges[i] = new std::bitset<8> [border_length];
		std::fill_n(edges[i], border_length, base);
		edge_costs[i] = new int * [border_length];
		overlap_costs[i] = new int * [border_length];
		for (int j = 0; j < border_length; ++j) {
			edge_costs[i][j] = new int[8];
			overlap_costs[i][j] = new int[8];
			std::fill_n(edge_costs[i][j], 8, BASE_EDGE_COST);
			std::fill_n(overlap_costs[i][j], 8, 0);
		}
	}
	/* desable edges for nodes along the border that go to nowhere */
	for (int i=0; i < border_length; ++i) {
		edges[0][i] &= 0b00011111;
		edges[border_length-1][i] &= 0b11110001;
		edges[i][0] &= 0b01111100;
		edges[i][border_length-1] &= 0b11000111;
	}
}

/* Destructor */
Mesh::~Mesh()
{
	for (int i=0; i < border_length; ++i) {
		delete [] edges[i];
		delete [] nodes[i];
		for (int j = 0; j < border_length; ++j) {
			delete [] edge_costs[i][j];
			delete [] overlap_costs[i][j];
		}
		delete [] edge_costs[i];
		delete [] overlap_costs[i];
	}
	delete [] edges;
	delete [] nodes;
	delete [] edge_costs;
	delete [] overlap_costs;
}

/* Adds a node to mesh */
void Mesh::add_node(Node n)
{
	if (!nodes[n.x][n.y]) {
		nodes[n.x][n.y] = true;
		num_nodes_++;
	}
}


/* Adds a edge to mesh */
void Mesh::add_edge(Edge e)
{
	assert_valid_node(e.start);
	assert_valid_node(e.end);
	assert_node_exist(e.start);
	assert_node_exist(e.end);
	int dir = get_direction(e);
	edges[e.start.x][e.start.y].set( dir );
	edges[e.end.x][e.end.y].set(opp_direction(dir));
	num_edges_++;
}


/* Removes a node */
void Mesh::remove_node(Node n)
{
	nodes[n.x][n.y] = false;
	num_nodes_--;
	if (n.x >= 1 && n.x < border_length-1 && n.y >= 1 && n.y < border_length-1) {
		remove_edge(Edge(n, Node(n.x, n.y-1)));
		remove_edge(Edge(n, Node(n.x+1, n.y-1)));
		remove_edge(Edge(n, Node(n.x+1, n.y)));
		remove_edge(Edge(n, Node(n.x+1, n.y+1)));
		remove_edge(Edge(n, Node(n.x, n.y+1)));
		remove_edge(Edge(n, Node(n.x-1, n.y+1)));
		remove_edge(Edge(n, Node(n.x-1, n.y)));
		remove_edge(Edge(n, Node(n.x-1, n.y-1)));
		return;
	}

	if (n.x >= 1) {
		remove_edge(Edge(n, Node(n.x-1, n.y)));
		if (n.y >= 1) {
			remove_edge(Edge(n, Node(n.x-1, n.y-1)));
			remove_edge(Edge(n, Node(n.x, n.y-1)));
		}
		if (n.y < border_length-1) {
			remove_edge(Edge(n, Node(n.x-1, n.y+1)));
			remove_edge(Edge(n, Node(n.x, n.y+1)));
		}
	}
	if (n.x < border_length-1) {
		remove_edge(Edge(n, Node(n.x+1, n.y)));
		if (n.y >= 1) {
			remove_edge(Edge(n, Node(n.x+1, n.y-1)));
			remove_edge(Edge(n, Node(n.x, n.y-1)));
		}
		if (n.y < border_length-1) {
			remove_edge(Edge(n, Node(n.x+1, n.y+1)));
			remove_edge(Edge(n, Node(n.x, n.y+1)));
		}
	}
}


/* Adds a cost bloom recursively starting at the cenre node */
void Mesh::add_cost_bloom(Node centre, int peak_cost, int dispersal_rate)
{
	const int xc[] = { 0, 1, 1, 1, 0, -1, -1, -1 };
	const int yc[] = { -1, -1, 0, 1, 1, 1, 0, -1 };
	Node t;
	for (int i = 0; i < 8; ++i) {
		t.x = centre.x + xc[i];
		t.y = centre.y + yc[i];
		if (is_valid_node(t)) {
			if (not is_node(t)) continue;
			if (edge_costs[centre.x][centre.y][i] < peak_cost) {
				edge_costs[centre.x][centre.y][i] = peak_cost;
				edge_costs[t.x][t.y][opp_direction(i)] = peak_cost;
			}
			if ( peak_cost - dispersal_rate > BASE_EDGE_COST) {
				add_cost_bloom(t, peak_cost - dispersal_rate, dispersal_rate);
			}
		}
	}
}


/* sets the cost of an edge */
void Mesh::set_edge_cost(const Edge e, int cost)
{
	int dir = get_direction(e);
	edge_costs[e.start.x][e.start.y][dir] = cost;
	edge_costs[e.end.x][e.end.y][opp_direction(dir)] = cost;
}


/* increases the cost of an edge by inc */
void Mesh::increase_edge_cost(const Edge e, int inc)
{
	int dir = get_direction(e);
	edge_costs[e.start.x][e.start.y][dir] += inc;
	edge_costs[e.end.x][e.end.y][opp_direction(dir)] += inc;
}

/* Adds an edge to the overlap cost map */
void Mesh::add_overlap_cost(const Edge e, int cost)
{
	int dir = get_direction(e);
	overlap_costs[e.start.x][e.start.y][dir] += cost;
	overlap_costs[e.end.x][e.end.y][opp_direction(dir)] += cost;
}


/* gets the cost of an edge */
int Mesh::edge_cost(const Edge e) const
{
	return edge_costs[e.start.x][e.start.y][get_direction(e)];
}

/* gets the cost of an edge includeing the overlaping bus route cost */
int Mesh::edge_and_overlap_cost(const Edge e) const
{
	int dir = get_direction(e);
	return edge_costs[e.start.x][e.start.y][dir] +
					overlap_costs[e.start.x][e.start.y][dir];
}


/* Removes and edge */
void Mesh::remove_edge(Edge e)
{
	assert_valid_node(e.start);
	assert_valid_node(e.end);
	int dir = get_direction(e);
	if (edges[e.start.x][e.start.y][dir]) {
		edges[e.start.x][e.start.y].reset(dir);
		edges[e.end.x][e.end.y].reset(opp_direction(dir));
		num_edges_--;
	}
}


/* returns the number of neighbors a node has */
int Mesh::num_neighbors(Node & n) const
{
	return edges[n.x][n.y].count();
}


/* Fills neighbors with up to 8 neighbors of node n */
void Mesh::get_neighbors(std::vector<Node> & neighbors, Node & n) const
{
	Edge e;
	e.start = n;
	e.end = n;
	const int xc[] = { 0, 1, 1, 1, 0, -1, -1, -1 };
	const int yc[] = { -1, -1, 0, 1, 1, 1, 0, -1 };
	for(int i = 0; i < 8; ++i) {
		e.end.x = n.x + xc[i];
		e.end.y = n.y + yc[i];
		if (not is_valid_node(e.end)) continue;
		if (is_edge(e)) {
			neighbors.push_back(e.end);
		}
	}
}

/* Returns if a node exists */
bool Mesh::is_node(const Node n) const
{
	assert_valid_node(n);
	return nodes[n.x][n.y];
}


/* returns wether is an edge */
bool Mesh::is_edge(const Edge e) const
{
	assert_valid_node(e.start);
	assert_valid_node(e.end);
	int dir = get_direction(e);

	return edges[e.start.x][e.start.y].test(dir) &&
				 edges[e.end.x][e.end.y].test(opp_direction(dir));
}


/***
 * Private functions 
 ***/


/* A simple switch to get the direction between two nodes 
 * DOES NOT CHECK IF THEY ARE BESIDE EACH OTHER */
int Mesh::get_direction(const Edge e) const
{
	switch(e.end.x - e.start.x) {
	case 1:
		return 2 + (e.end.y - e.start.y);
	case 0:
		return 2 + 2 * (e.end.y - e.start.y);
	case -1:
		return 6 - (e.end.y - e.start.y);
	default:
		std::cerr << "Directional error" << std::endl;
		exit(1);
	}
}


/* returns the opposite direction of the one entered */
int Mesh::opp_direction(int dir) const
{
	return (4+dir)%8;
}


/* Checks if a node could exist */
bool Mesh::is_valid_node(const Node n) const
{
	return	n.x >= 0 and n.x < border_length and
					n.y >= 0 and n.y < border_length;
}


void Mesh::assert_valid_node(const Node n) const
{
	assert(n.x >= 0 && n.x < border_length);
	assert(n.y >= 0 && n.y < border_length);
}


void Mesh::assert_node_exist(const Node n) const
{
	if (not nodes[n.x][n.y]) {
		std::cout << n;
	}
	assert(nodes[n.x][n.y]);
}

/** Print Function **/
std::ostream& operator<<(std::ostream& out, Mesh const& m)
{
	int MAX = m.border_length;
	for (int y=0; y< MAX; ++y) {
		for (int x=0; x < MAX; ++x) {
			if (y and x) {
				/* make X */
				bool n_di = m.is_edge(Edge(Node(x-1,y-1), Node(x,y)));
				bool p_di = m.is_edge(Edge(Node(x-1,y), Node(x,y-1)));
				if (p_di and n_di) {
					out << 'X';
				} else if (p_di) {
					out << '/';
				} else if (n_di) {
					out << '\\';
				} else {
					out << ' ';
				}
			}
			if (y) {
				if(m.is_edge(Edge(Node(x,y-1), Node(x,y)))) {
					out << '|';
				} else {
					out << ' ';
				}
			}
		}
		out << std::endl;

		for (int x=0; x< MAX; ++x) {
			if (x) {
				if (m.is_edge(Edge(Node(x-1,y), Node(x,y)))) {
					out <<'-';
				} else {
					out << ' ';
				}
			}
			if (m.nodes[x][y]) {
				out << 'O';
			} else {
				out << ' ';
			}
		}
		out << std::endl;	
		}
	return out;
}


