/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Marcus Karpoff
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _MESH_H_
#define _MESH_H_

#include <array>
#include <vector>
#include <bitset>

#include "node.h"
#include "edge.h"

extern int BASE_EDGE_COST;
extern int MAX_EDGE_COST;
extern int OVERLAP_COST;

class Mesh {
	public:
		const int border_length;
		bool ** nodes;
		std::bitset<8> ** edges;
		int *** edge_costs;
		int *** overlap_costs;

		Mesh(int border_length, int base=0b01010101);
		~Mesh();
		void add_node(Node n);
		void add_edge(Edge e);
		void add_cost_bloom(Node centre, int peak_cost, int dispersal_rate=1);
		int edge_cost(const Edge e) const;
		int edge_and_overlap_cost(const Edge e) const;
		void set_edge_cost(const Edge e, int cost);
		void increase_edge_cost(const Edge e, int inc);
		void add_overlap_cost(const Edge e, int cost = OVERLAP_COST);
		void remove_node(Node n);
		void remove_edge(Edge e);
		int num_neighbors(Node & n) const;
		void get_neighbors(std::vector<Node> & neighbors, Node & n) const;
		bool is_node(const Node n) const;
		bool is_edge(const Edge e) const;
		bool is_valid_node(const Node n) const;
		inline int num_nodes() const { return num_nodes_; };
		inline int num_edges() const { return num_edges_; };

	private:
		int num_nodes_;
		int num_edges_;
		int get_direction(const Edge e) const;
		int opp_direction(int dir) const;
		void assert_valid_node(const Node n) const;
		void assert_node_exist(const Node n) const;
};

std::ostream& operator<<(std::ostream& out, Mesh const& m);
#endif
